--[[

Game:	Farming Simulator 19
Title:	UnitConvert Lite (FillLevelsDisplay)
Author:	ThundR

--]]

-- [Declarations] ***********************************************************************************************************************************

UnitExtension.FillLevelsDisplay = {}

-- [Initialization] *********************************************************************************************************************************

function UnitExtension:initializeFillLevelsDisplay()
-- Initializes FillLevelsDisplay

	-- FillLevelsDisplay additional functions
	HMUtils.setFunction(self, "FillLevelsDisplay", "getYieldFillTypes", false)

	-- FillLevelsDisplay overrides
	HMUtils.setFunction(self, "FillLevelsDisplay", "new")
	HMUtils.setFunction(self, "FillLevelsDisplay", "draw")
	HMUtils.setFunction(self, "FillLevelsDisplay", "setVehicle")
	HMUtils.setFunction(self, "FillLevelsDisplay", "storeScaledValues")
	HMUtils.setFunction(self, "FillLevelsDisplay", "updateFillLevelBuffers")
	HMUtils.setFunction(self, "FillLevelsDisplay", "updateFillLevelFrames")
end

-- [FillLevelsDisplay Additional Functions] *********************************************************************************************************

function UnitExtension.FillLevelsDisplay:getYieldFillTypes(vehicle)
-- Returns the yield values of current vehicle and all attachments
	-- [vehicle]: Vehicle to inspect

	if HMUtils.argIsInvalid(HMUtils.isTable(vehicle), "vehicle") then
		return
	end

	if vehicle.spec_attachable then
		vehicle = Utils.getNoNil(vehicle:getRootVehicle(), vehicle)
	end

	local yieldFillTypes = Utils.getNoNil(vehicle.unitExtension.yieldFillTypes, {})

	if vehicle.spec_attacherJoints then
		local attachedImplements = vehicle:getAttachedImplements()
		for _, implement in pairs(attachedImplements) do
			local object = implement.object
			if object.unitExtension.yieldFillTypes then
				for fillType, yield in pairs(object.unitExtension.yieldFillTypes) do
					if yieldFillTypes[fillType] then
						yieldFillTypes[fillType] = (yieldFillTypes[fillType] + yield) / 2 -- average
					else
						yieldFillTypes[fillType] = yield
					end
				end
			end
		end
	end

	return yieldFillTypes
end

-- [FillLevelsDisplay Overrides] ********************************************************************************************************************

function UnitExtension.FillLevelsDisplay.new(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "FillLevelsDisplay.new")
	local ret = HMUtils.pack(savedFunc(...))

	if ret.n > 0 then
		local self = ret[1]
		self.fillTypeUnitBuffer = {}
	end

	return HMUtils.unpack(ret)
end

function UnitExtension.FillLevelsDisplay:draw(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "FillLevelsDisplay.draw")
	local ret = HMUtils.pack(savedFunc(self, ...))

	if self:getVisible() then
		local baseX, baseY = self:getPosition()
		local width = self:getWidth()

		for i = 1, #self.fillLevelBuffer do
			if self.fillTypeUnitBuffer[i] then
				local posX = (baseX + width) - self.fillLevelTextOffsetX
				local posY = baseY + (self.frameHeight * (i-1)) + (self.fillLevelTextOffsetY * 1.75)
				renderText(posX, posY, self.fillLevelTextSize, self.fillTypeUnitBuffer[i])
			end
		end
	end

	return HMUtils.unpack(ret)
end

function UnitExtension.FillLevelsDisplay:setVehicle(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "FillLevelsDisplay.setVehicle")
	local ret = HMUtils.pack(savedFunc(self, ...))

	if self.vehicle then
		self.rootVehicle = self.vehicle:getRootVehicle()
	else
		self.rootVehicle = nil
	end

	return HMUtils.unpack(ret)
end

function UnitExtension.FillLevelsDisplay:storeScaledValues(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "FillLevelsDisplay.storeScaledValues")
	local ret = HMUtils.pack(savedFunc(self, ...))

	local aspectRatio = getScreenAspectRatio()
	local scaleFactor = aspectRatio * 0.45

	self.frameHeight = self:scalePixelToScreenVector(FillLevelsDisplay.SIZE.FILL_TYPE_FRAME) * scaleFactor

	return HMUtils.unpack(ret)
end

function UnitExtension.FillLevelsDisplay:updateFillLevelBuffers(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "FillLevelsDisplay.updateFillLevelBuffers")

	self.fillTypeUnitBuffer = {}

	local ret = HMUtils.pack(savedFunc(self, ...))

	return HMUtils.unpack(ret)
end

function UnitExtension.FillLevelsDisplay:updateFillLevelFrames(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "FillLevelsDisplay.updateFillLevelFrames")

	for i, fillLevelInfo in pairs(self.fillLevelBuffer) do
		local unitText = ""
		local unitSet = "VEHICLE"
		local vehicle = self.vehicle
		local rootVehicle = self.rootVehicle

		if rootVehicle and rootVehicle.getUnitSet then
			unitSet = rootVehicle:getUnitSet()
		end

		local fillTypeIndex = Utils.getNoNil(fillLevelInfo.realFillType, fillLevelInfo.fillType)
		local fillLevel = tonumber(fillLevelInfo.fillLevel)
		local capacity = tonumber(fillLevelInfo.capacity)
		local fillTypeUnit, conversionFactor = g_ucLite:getFillTypeUnit(fillTypeIndex, unitSet)

		fillLevelInfo.fillTypeUnit = {
			index  = fillTypeUnit.index,
			factor = conversionFactor
		}

		if fillTypeIndex ~= FillType.UNKNOWN then
			unitText = HMUtils.properCase(fillTypeUnit.title)
		end

		if fillLevel and fillLevel > 0 and fillLevel < math.huge then
			fillLevelInfo.fillLevel = fillLevel * conversionFactor
		end

		if capacity and capacity > 0 and capacity < math.huge then
			fillLevelInfo.capacity = capacity * conversionFactor
		end

		local yieldFillTypes = self:getYieldFillTypes(vehicle)
		if yieldFillTypes then
			local yield = yieldFillTypes[fillTypeIndex]
			if yield and yield > 0 then
				local areaFactor = 10000 / g_i18n:getArea(g_ucLite.areaScale)

				yield = yield * conversionFactor * areaFactor
				yield = HMUtils.numToStr(yield, fillTypeUnit.precision, "ROUND")
				unitText = string.format("%s %s/%s", yield, fillTypeUnit.titleShort, g_i18n:getAreaUnit())
			end
		end

		table.insert(self.fillTypeUnitBuffer, unitText)
	end

	local ret = HMUtils.pack(savedFunc(self, ...))

	for i, fillLevelInfo in pairs(self.fillLevelBuffer) do
		local fillLevel = tonumber(fillLevelInfo.fillLevel)
		local capacity = tonumber(fillLevelInfo.capacity)
		if fillLevel and capacity and capacity > 0 then
			local fillPct = math.floor((fillLevel / capacity) * 100)
			local fillTypeUnit = g_ucLite:getUnit(fillLevelInfo.fillTypeUnit.index)
			fillLevel = HMUtils.numToStr(fillLevel, fillTypeUnit.precision)

			if self.fillLevelTextBuffer[i] then
				self.fillLevelTextBuffer[i] = string.format("%s (%s%%)", fillLevel, fillPct)
			end
		end
	end

	return HMUtils.unpack(ret)
end
