--[[

Game:	Farming Simulator 19
Title:	UnitConvert Lite
Author:	ThundR

Allows Farming Simulator 19 to display different units of measurement
Slimmed version of the UnitExtension (possibly DisplayExtension) app for HarvestMaster

TODO:
	- Add the toggle inputBinding for switching units (In Progress --- think of something besides currentVehicleUnit)
	- Create conversion function(s)

--]]

-- [Declarations] ***********************************************************************************************************************************

g_modName = g_currentModName
g_modDesc = g_modManager:getModByName(g_modName)
g_modPath = g_modDesc.modDir
g_modFile = g_modDesc.modFile

UnitExtension = {}

-- L10n Constant(s)
UnitExtension.L10N_PREFIX = "ue_"

-- Message Constants
UnitExtension.MSG = {
	UNIT_EXISTS  			= "Unit %q already registered",
	INVALID_UNIT			= "Invalid unit",
	FILLTYPE_EXISTS			= "FillType %q already registered",
	FILLTYPE_GROUP_EXISTS	= "FillType group %q already registered",
	INVALID_CONFIG_VERSION 	= "Invalid configuration version, current is v%s\n%s" -- includes current config version and filename
}

-- Override Tables
UnitExtension.Mission00 = {}
UnitExtension.FillTypeManager = {}
UnitExtension.AnimalFoodManager = {}

g_ucLite = {}
g_ucLite.name = "unitExtension"
g_ucLite.className = "UnitExtension"
g_ucLite.debugLevel = 0

function g_ucLite:class()
	return UnitExtension
end

setmetatable(g_ucLite, {__index = UnitExtension})
getfenv(0).g_ucLite = g_ucLite

-- [Extra Source Files] *****************************************************************************************************************************

source(g_modPath.."scripts/Utils.lua")
source(g_modPath.."scripts/Dialogs.lua")
source(g_modPath.."scripts/Vehicles.lua")
source(g_modPath.."scripts/InGameMenu.lua")
source(g_modPath.."scripts/SpeedMeterDisplay.lua")
source(g_modPath.."scripts/FillLevelsDisplay.lua")
--source(g_modPath.."scripts/FieldInfoDisplay.lua") -- unused (for now)

-- [Events] *****************************************************************************************************************************************

function UnitExtension:onInitialize(...)
	self:initializeUnits()
	self:initializeFillTypeGroups()
	self:initializeFillTypes()
	self:initializeSettings()

	-- FillType event listener(s)
	HMUtils.setFunction(self, "FillTypeManager", "loadMapData")

	-- FoodGroup event listener(s)
	HMUtils.setFunction(self, "AnimalFoodManager", "loadFoodGroups")

	-- Initialize InGameMenu features
	self:initializeInGameMenu()

	-- Initialize Dialog features
	self:initializeDialogs()

	-- Initialize FillLevelsDisplay features
	self:initializeFillLevelsDisplay()

	-- Initialize SpeedMeterDisplay features
	self:initializeSpeedMeterDisplay()

	-- Initialize FieldInfoDisplay features
	--self:onInitMapFieldInfoDisplay() -- Not currently used

	-- Initialize Vehicle features
	self:initializeVehicles()
end

function UnitExtension:onLoadMapFinished(...)
	self:loadFillTypes()
	self:loadConfig()

	if DebugTools then
		--DebugTools:printTable(self, 4, true) -- Check internal tables
	end
end

-- [FillTypeManager Overrides] **********************************************************************************************************************

function UnitExtension.FillTypeManager:loadMapData(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "FillTypeManager.loadMapData")
	local ret = HMUtils.pack(savedFunc(self, ...))

	g_ucLite:preLoadFillTypes()

	return HMUtils.unpack(ret)
end

-- [AnimalFoodManager Overrides] ********************************************************************************************************************

function UnitExtension.AnimalFoodManager:loadFoodGroups(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "AnimalFoodManager.loadFoodGroups")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(self, ...))

	local xmlFile = HMUtils.getRealArg(args, 1)
	if xmlFile then
	end

	return HMUtils.unpack(ret)
end

-- [Initialization] *********************************************************************************************************************************

function UnitExtension:initializeUnits()
-- Initializes unit tables

	self.unitTypesByIndex = {}
	self.unitTypesByName = {}
	self.unitType = {}

	table.insert(self.unitTypesByIndex, {name = "VOLUME"})
	table.insert(self.unitTypesByIndex, {name = "WEIGHT"})
	table.insert(self.unitTypesByIndex, {name = "PIECE"})
	table.insert(self.unitTypesByIndex, {name = "LENGTH"})

	for i, unitType in ipairs(self.unitTypesByIndex) do
		unitType.index = i
		self.unitType[unitType.name] = unitType.index
		self.unitTypesByName[unitType.name] = unitType
	end

	self.unitSetsByIndex = {}
	self.unitSetsByName = {}
	self.unitSet = {}

	self.numUnitSets = {
		VEHICLE = 2,
	}

	self.currentUnitSet = {
		VEHICLE = 1,
	}

	for i = 1, self.numUnitSets.VEHICLE do
		table.insert(self.unitSetsByIndex, {name = "VEHICLE"..tostring(i)})
	end
	table.insert(self.unitSetsByIndex, {name = "ANIMAL"})
	table.insert(self.unitSetsByIndex, {name = "PRICE"})

	for i, unitSet in ipairs(self.unitSetsByIndex) do
		unitSet.index = i
		self.unitSet[unitSet.name] = unitSet.index
		self.unitSetsByName[unitSet.name] = unitSet
	end

	self.unitsByIndex = {}
	self.unitsByName = {}
	self.unit = {}

	-- Default units
	self:registerUnit("LITER",	1, 0, "VOLUME", nil, nil, true) -- locked (true)
	self:registerUnit("PIECE",	1, 0, "PIECE",  nil, nil, true) -- locked (true)
	self:registerUnit("ANIMAL",	1, 0, "PIECE",  nil, nil, true) -- locked (true)
	self:registerUnit("BALE",	1, 0, "PIECE",  nil, nil, true) -- locked (true)
	self:registerUnit("TONNE",	1, 0, "WEIGHT", nil, nil, true) -- locked (true)

end

function UnitExtension:initializeFillTypeGroups()
-- Initializes fillType group tables

	self.fillTypeGroupsByIndex = {}
	self.fillTypeGroupsByName = {}
	self.fillTypeGroup = {}

	self:registerFillTypeGroup("GRAIN",  nil, nil,  "LITER")
	self:registerFillTypeGroup("LIQUID", nil, nil,  "LITER")
	self:registerFillTypeGroup("ANIMAL", nil, true, "ANIMAL")	-- locked (true)
	self:registerFillTypeGroup("PIECE",  nil, true, "PIECE")	-- locked (true)
	self:registerFillTypeGroup("BALE",   nil, true, "BALE")		-- locked (true)
	self:registerFillTypeGroup("OTHER",  nil, nil,  "LITER")
	self:registerFillTypeGroup("MISC",   nil, true, "LITER")	-- locked (true)
end

function UnitExtension:initializeFillTypes()
-- Initializes fillType tables

	self.fillTypesByIndex = {}
	self.fillTypesByName = {}
	self.fillType = {}
end

function UnitExtension:initializeSettings()
-- Initializes UnitExtension settings

	self.areaScale = 1.0
end

-- [Loading] ****************************************************************************************************************************************

function UnitExtension:preLoadFillTypes()
-- Pre-Loads fillType tables (get default values)

	for _, fillType in pairs(g_fillTypeManager.fillTypes) do
		self:registerFillType(fillType.name, nil, true) -- unscale fillTypes loaded via xml (true)
	end
end

function UnitExtension:loadFillTypes()
-- Loads fillType tables

	for _, fillType in pairs(g_fillTypeManager.fillTypes) do
		local newFillType = self:getFillType(fillType.name)
		if not newFillType then
			newFillType = self:registerFillType(fillType.name) -- register fiilTypes created by other mods (don't unscale these)
		end

		-- Update masses set by mods (real weight) and try to ignore zero weight mods
		if fillType.massPerLiter ~= newFillType.defaults.oldMassPerLiter then -- Giants mass has changed since registration
			if fillType.massPerLiter >= 0.000001 then -- try to ignore zero weight
				self:setFillTypeMass(fillType.name, fillType.massPerLiter)
			end
		end
	end

	-- Set OTHER group fillTypes
	local otherFillTypes = g_fillTypeManager:getFillTypesByCategoryNames("BULK")
	for _, fillTypeIndex in pairs(otherFillTypes) do
		self:setFillTypeGroup(fillTypeIndex, "OTHER")
	end

	-- Set GRAIN group fillTypes
	local grainFillTypes = g_fillTypeManager:getFillTypesByCategoryNames("COMBINE")
	for _, fillTypeIndex in pairs(grainFillTypes) do
		self:setFillTypeGroup(fillTypeIndex, "GRAIN")
	end

	-- Set LIQUID group fillTypes
	local liquidFillTypes = g_fillTypeManager:getFillTypesByCategoryNames("LIQUID SPRAYER SLURRYTANK")
	for _, fillTypeIndex in pairs(liquidFillTypes) do
		self:setFillTypeGroup(fillTypeIndex, "LIQUID")
	end

	-- Set PIECE group fillTypes
	local pieceFillTypes = g_fillTypeManager:getFillTypesByCategoryNames("PIECE")
	for _, fillTypeIndex in pairs(pieceFillTypes) do
		self:setFillTypeGroup(fillTypeIndex, "PIECE")
	end

	-- Set ANIMAL group fillTypes
	local animalFillTypes = g_fillTypeManager:getFillTypesByCategoryNames("ANIMAL HORSE")
	for _, fillTypeIndex in pairs(animalFillTypes) do
		self:setFillTypeGroup(fillTypeIndex, "ANIMAL")
	end

	-- Add all "bale" fillTypes to the BALE group
	for _, fillType in pairs(g_fillTypeManager.fillTypes) do
		if string.find(fillType.name, "BALE") then
			self:setFillTypeGroup(fillType.name, "BALE")
		end
	end

	-- Special fillType group settings
	self:setFillTypeGroup("SEEDS",			"GRAIN")
	self:setFillTypeGroup("OILSEEDRADISH",  "OTHER")
	self:setFillTypeGroup("POPLAR",			"OTHER")
	self:setFillTypeGroup("COTTON",			"OTHER")
	self:setFillTypeGroup("GRASS",			"OTHER")
	self:setFillTypeGroup("WOOL",			"OTHER")
end

-- [Configuration] **********************************************************************************************************************************

function UnitExtension:loadConfig()
-- Loads configuration information

	self.config =  {}
	self.config.version = 2
	self.config.filename = "ucLite.xml"
	self.config.baseXML = Utils.getFilename("config.xml", g_modPath)
	self.config.userXML = Utils.getFilename(self.config.filename, getUserProfileAppPath())

	-- Add savegame configuration (if found)
	self.config.savegameDirectory = g_currentMission.missionInfo.savegameDirectory
	if HMUtils.isString(self.config.savegameDirectory) then
		self.config.savegameDirectory = self.config.savegameDirectory.."/"
		self.config.saveXML = Utils.getFilename(self.config.filename, self.config.savegameDirectory)
	end

	-- Add map configuration (if found)
	local mapInfo = HMUtils.getTable(g_currentMission, "missionInfo.map")
	if mapInfo then
		if mapInfo.isModMap == true then
			self.config.mapXML = Utils.getFilename(self.config.filename, mapInfo.baseDirectory)
		end
	end

	local configFile = self.config.baseXML
	if HMUtils.isString(self.config.mapXML) and fileExists(self.config.mapXML) then
		configFile = self.config.mapXML
	elseif HMUtils.isString(self.config.saveXML) and fileExists(self.config.saveXML) then
		configFile = self.config.saveXML
	elseif HMUtils.isString(self.config.userXML) and fileExists(self.config.userXML) then
		configFile = self.config.userXML
	end

	self:loadFromXML(configFile, self.className)
end

function UnitExtension:loadFromXML(configFile, baseKey)
-- Loads configuration information from configuration file(s)

	if HMUtils.argIsInvalid(HMUtils.isString(configFile), "configFile")
	or HMUtils.argIsInvalid(HMUtils.isString(baseKey), "baseKey")
	then
		return
	end

	if not fileExists(configFile) then
		HMUtils.displayMsg(self:getMsg("FILE_NOT_FOUND"), configFile)

		if configFile ~= self.config.baseXML then
			self:loadFromXML(self.config.baseXML, baseKey)
		end

		return
	end

	local xmlFile = loadXMLFile(baseKey, configFile)
	if not xmlFile or not hasXMLProperty(xmlFile, baseKey) then
		HMUtils.displayMsg(self:getMsg("INVALID_XML_FILE"), configFile)

		if configFile ~= self.config.baseXML then
			self:loadFromXML(self.config.baseXML, baseKey)
		end

		delete(xmlFile)
		return
	end

	-- Load general information
	local configVersion = Utils.getNoNil(getXMLInt(xmlFile, baseKey.."#version"), 1)
	if configVersion ~= self.config.version then
		HMUtils.displayMsg(self:getMsg("INVALID_CONFIG_VERSION"), self.config.version, configFile)

		if configFile ~= self.config.baseXML then
			self:loadFromXML(self.config.baseXML, baseKey)
		end

		delete(xmlFile)
		return
	end

	-- Load settings
	local settingsKey = string.format("%s.settings", baseKey)

	-- Optionally activate debugger
	local debugMode = getXMLString(xmlFile, settingsKey.."#debugMode")
	debugMode = HMUtils.upperCase(debugMode)
	if HMUtils.DEBUG[debugMode] then
		HMUtils.displayMsg("Global debug mode has been set to %s", debugMode)
		HMUtils.globalDebugMode = HMUtils.DEBUG[debugMode]
	end

	-- Optionally set area scale
	local areaScale = getXMLFloat(xmlFile, settingsKey.."#areaScale")
	if areaScale and areaScale > 0 then
		HMUtils.displayMsg("Area scale has been set to %s", areaScale)
		self.areaScale = areaScale
	end

	-- Load unit information
	local idx = 0
	while true do
		local unitKey = string.format("%s.units.unit(%d)", baseKey, idx)
		if not hasXMLProperty(xmlFile, unitKey) then
			break
		end

		local unitInfo = {
			name 	   = getXMLString(xmlFile, unitKey.."#name"),
			factor 	   = getXMLFloat(xmlFile, unitKey.."#factor"),
			precision  = getXMLInt(xmlFile, unitKey.."#precision"),
			type	   = getXMLString(xmlFile, unitKey.."#type"),
			title	   = getXMLString(xmlFile, unitKey.."#title"),
			titleShort = getXMLString(xmlFile, unitKey.."titleShort")
		}

		if not unitInfo.name then
			HMUtils.displayMsg(self:getMsg("INVALID_XML_VALUE_AT_INDEX"), "<unit> name", idx, configFile)
		else
			local unit = self:getUnit(unitInfo.name)
			if unit then
				if not unit.locked then
					self:setUnitFactor(unit.name, unitInfo.factor)
					self:setUnitPrecision(unit.name, unitInfo.precision)
				end
				self:setUnitTitle(unit.name, unitInfo.title, unitInfo.titleShort)
			else
				local unit = self:registerUnit(unitInfo.name, unitInfo.factor, unitInfo.precision, unitInfo.type, unitInfo.title, unitInfo.titleShort)
				if unit then
					unit.isUser = true
				end
			end
		end

		idx = idx + 1
	end

	-- Load fillType group information
	local idx = 0
	while true do
		local groupKey = string.format("%s.fillTypeGroups.fillTypeGroup(%d)", baseKey, idx)
		if not hasXMLProperty(xmlFile, groupKey) then
			break
		end

		local groupInfo = {
			name	= getXMLString(xmlFile, groupKey.."#name"),
			title	= getXMLString(xmlFile, groupKey.."#title"),
			units	= {}
		}

		if not groupInfo.name then
			HMUtils.displayMsg(self:getMsg("INVALID_XML_VALUE_AT_INDEX"), "<fillTypeGroup> name", idx, configFile)
		else
			local unitsKey = string.format("%s.units", groupKey)
			if hasXMLProperty(xmlFile, unitsKey) then
				for i, unitSet in ipairs(self.unitSetsByIndex) do
					local xmlName = HMUtils.lowerCase(unitSet.name)
					groupInfo.units[i] = getXMLString(xmlFile, unitsKey.."#"..xmlName)
				end
			end

			local group = self:getFillTypeGroup(groupInfo.name)
			if group then
				self:setFillTypeGroupTitle(groupInfo.name, groupInfo.title)
				if not group.locked then
					for i, unitSet in ipairs(self.unitSetsByIndex) do
						self:setFillTypeGroupUnits(group.name, i, groupInfo.units[i])
					end
				end
			else
				local group = self:registerFillTypeGroup(groupInfo.name, groupInfo.title, nil, groupInfo.units)
				if group then
					group.isUser = true
				end
			end
		end

		idx = idx + 1
	end

	-- Load fillType information
	local idx = 0
	while true do
		local fillTypeKey = string.format("%s.fillTypes.fillType(%d)", baseKey, idx)
		if not hasXMLProperty(xmlFile, fillTypeKey) then
			break
		end

		local fillTypeInfo = {
			name			 = getXMLString(xmlFile, fillTypeKey.."#name"),
			group			 = getXMLString(xmlFile, fillTypeKey.."group"),
			mass			 = getXMLFloat(xmlFile, fillTypeKey.."#mass"),
			showOnPriceTable = getXMLBool(xmlFile, fillTypeKey.."#showOnPriceTable")
		}

		local fillType = self:getFillType(fillTypeInfo.name)
		if fillType then
			self:setFillTypeGroup(fillType.name, fillTypeInfo.group)
			if fillTypeInfo.mass then
				fillTypeInfo.mass = fillTypeInfo.mass / 1000000 -- convert to mt/l
				self:setFillTypeMass(fillType.name, fillTypeInfo.mass)
			end
			self:setFillTypeShowOnPriceTable(fillType.name, fillTypeInfo.showOnPriceTable)
		end

		idx = idx + 1
	end

	delete(xmlFile)

	if configFile == self.config.baseXML then
		self:saveToXML(self.config.userXML, baseKey)
	end
end

function UnitExtension:saveToXML(configFile, baseKey)
-- Saves configuration to external file

	if HMUtils.argIsInvalid(HMUtils.isString(configFile), "configFile")
	or HMUtils.argIsInvalid(HMUtils.isString(baseKey), "baseKey")
	then
		return
	end

	local xmlFile = createXMLFile(baseKey, configFile, baseKey)
	if not xmlFile then
		HMUtils.displayMsg(self:getMsg("FAILED_TO_CREATE_XML_FILE"), configFile)
		return
	end

	-- Save version information
	setXMLInt(xmlFile, baseKey.."#version", self.config.version)

	-- Save unit information
	local idx = 0
	for _, unit in ipairs(self.unitsByIndex) do
		local unitKey = string.format("%s.units.unit(%d)", baseKey, idx)
		local unitType = self:getUnitType(unit.type)
		if not unit.locked then
			setXMLString(xmlFile, unitKey.."#name", unit.name)
			setXMLFloat(xmlFile, unitKey.."#factor", unit.factor)
			setXMLInt(xmlFile, unitKey.."#precision", unit.precision)
			setXMLString(xmlFile, unitKey.."#type", unitType.name)

			if unit.hasUserTitle then
				setXMLString(xmlFile, unitKey.."#title", unit.title)
			end
			if unit.hasUserTitleShort then
				setXMLString(xmlFile, unitKey.."#titleShort", unit.titleShort)
			end

			idx = idx + 1
		end
	end

	-- Save fillType group information
	local idx = 0
	for _, group in ipairs(self.fillTypeGroupsByIndex) do
		local groupKey = string.format("%s.fillTypeGroups.fillTypeGroup(%d)", baseKey, idx)
		if not group.locked then
			setXMLString(xmlFile, groupKey.."#name", group.name)
			if group.hasUserTitle then
				setXMLString(xmlFile, groupKey.."#title", group.title)
			end

			local unitsKey = string.format("%s.units", groupKey)
			for i, unitSet in ipairs(self.unitSetsByIndex) do
				local unit = self:getUnit(group.units[i])
				local xmlName = HMUtils.lowerCase(unitSet.name)
				setXMLString(xmlFile, unitsKey.."#"..xmlName, unit.name)
			end

			idx = idx + 1
		end
	end

	-- Save fillType information
	local idx = 0
	local fillTypeList = HMUtils.sortTableByKey(self.fillTypesByIndex, "group") -- TODO: Take a better look at the sorting function(s)
	for _, fillType in ipairs(fillTypeList) do
		local fillTypeKey = string.format("%s.fillTypes.fillType(%d)", baseKey, idx)
		local fillTypeGroup = self:getFillTypeGroup(fillType.group)
		if not fillType.hidden and not fillTypeGroup.locked then
			setXMLString(xmlFile, fillTypeKey.."#name", fillType.name)
			setXMLString(xmlFile, fillTypeKey.."#group", fillTypeGroup.name)

			if fillType.massPerLiter ~= fillType.defaults.massPerLiter
			and fillType.massPerLiter ~= fillType.parent.massPerLiter
			then -- mass has been changed
				local mass = fillType.massPerLiter * 1000000 -- convert to kg/m³
				setXMLFloat(xmlFile, fillTypeKey.."#mass", mass)
			end

			idx = idx + 1
		end
	end

	saveXMLFile(xmlFile)
	delete(xmlFile)
end

-- [Base Functions] *********************************************************************************************************************************

function UnitExtension:getLocalText(text, ...)
-- Returns a local (overridden) i18n text value
	-- [text]:  Text key (string)
	-- [...]:	Variables passed to string.format function (any) > Optional

	return HMUtils.getText(UnitExtension.L10N_PREFIX..text, ...)
end

function UnitExtension:getMsg(msg)
-- Returns a saved log message
-- Calls HMUtils.getMsg() if not found
	-- [msg]: Message to retrieve

	if not HMUtils.isString(msg) then
		return
	end

	msg = HMUtils.upperCase(msg)
	return Utils.getNoNil(UnitExtension.MSG[msg], HMUtils.getMsg(msg))
end

-- [Unit Functions] *********************************************************************************************************************************

function UnitExtension:registerUnit(name, factor, precision, unitType, title, titleShort, isLocked)
-- Registers and returns a new unit
	-- [name]:		 Unit name
	-- [factor]:	 Unit conversion factor > See UnitExtension:setUnitFactor()
	-- [precision]:  Unit precision > See UnitExtension:setUnitPrecsion()
	-- [unitType]:	 Unit type (name or index) > See UnitExtension:setUnitType()
	-- [title]:		 Unit title > See UnitExtension:setUnitTitle()
	-- [titleShort]: Unit short > See UnitExtension:setUnitTitle()
	-- [isLocked]:	 Prevent certain unit values from being changed > Default = false

	unitType = Utils.getNoNil(unitType, self.unitType.VOLUME)
	isLocked = Utils.getNoNil(isLocked, false)

	unitType = self:getUnitType(unitType)

	if HMUtils.argIsInvalid(HMUtils.isString(name), "name")
	or HMUtils.argIsInvalid(unitType, "unitType")
	or HMUtils.argIsInvalid(HMUtils.isBoolean(isLocked), "isLocked")
	then
		return
	end

	local unitInfo = self:getUnit(name)
	if HMUtils.errorOnTrue(unitInfo, self:getMsg("UNIT_EXISTS"), name) then
		return unitInfo
	end

	unitInfo = {
		name	 = HMUtils.upperCase(name),
		type	 = unitType.index,
		locked	 = isLocked,
		isUser	 = false, -- this is set in the load xml on registration
		defaults = {}
	}
	table.insert(self.unitsByIndex, unitInfo)
	unitInfo.index = #self.unitsByIndex
	self.unitsByName[unitInfo.name] = unitInfo
	self.unit[unitInfo.name] = unitInfo.index

	self:setUnitFactor(unitInfo.name, factor)
	self:setUnitPrecision(unitInfo.name, precision)
	self:setUnitTitle(unitInfo.name, title, titleShort)

	HMUtils.displayMsg("Unit %q registered successfully", unitInfo.name)
	return unitInfo
end

function UnitExtension:setUnitFactor(unit, factor)
-- Sets specified unit factor value
	-- [unit]: 	 Target unit (name or index)
	-- [factor]: Unit factor > Default = 1

	unit = self:getUnit(unit)
	if HMUtils.argIsInvalid(unit, "unit") then
		return
	end

	local oldFactor = Utils.getNoNil(unit.factor, 1)
	local newFactor = Utils.getNoNil(factor, oldFactor)

	if HMUtils.argIsInvalid(HMUtils.isNumber(newFactor) and newFactor > 0, "factor") then
		newFactor = oldFactor
	end

	unit.defaults.factor = Utils.getNoNil(unit.defaults.factor, newFactor)
	unit.factor = newFactor
end

function UnitExtension:setUnitPrecision(unit, precision)
-- Sets specified unit precision value
	-- [unit]: 	 	Target unit (name or index)
	-- [precision]: Unit precision > Default = 0

	unit = self:getUnit(unit)
	if HMUtils.argIsInvalid(unit, "unit") then
		return
	end

	local oldPrecision = Utils.getNoNil(unit.precision, 0)
	local newPrecision = Utils.getNoNil(precision, oldPrecision)

	if HMUtils.argIsInvalid(HMUtils.isNumber(newPrecision) and newPrecision >= 0, "precision") then
		newPrecision = oldPrecision
	end

	unit.defaults.precision = Utils.getNoNil(unit.defaults.precision, newPrecision)
	unit.precision = newPrecision
end

function UnitExtension:setUnitTitle(unit, title, titleShort)
-- Sets specified unit title and abbreviated title values
	-- [unit]: 		 Target unit (name or index)
	-- [title]:		 Unit title
	-- [titleShort]: Unit abbreviated title

	unit = self:getUnit(unit)
	if HMUtils.argIsInvalid(unit, "unit") then
		return
	end

	local l10n_title = UnitExtension.L10N_PREFIX.."unit_"..HMUtils.lowerCase(unit.name)
	local l10n_titleShort = l10n_title.."Short"

	l10n_title = HMUtils.getText(l10n_title)
	l10n_titleShort = HMUtils.getText(l10n_titleShort)

	local oldTitle = Utils.getNoNil(unit.title, l10n_title)
	local oldTitleShort = Utils.getNoNil(unit.titleShort, l10n_titleShort)

	local newTitle = Utils.getNoNil(title, oldTitle)
	local newTitleShort = Utils.getNoNil(titleShort, oldTitleShort)

	if HMUtils.argIsInvalid(HMUtils.isString(newTitle), "title") then
		newTitle = oldTitle
	end
	if HMUtils.argIsInvalid(HMUtils.isString(newTitleShort), "titleShort") then
		newTitleShort = oldTitleShort
	end

	unit.defaults.title = Utils.getNoNil(unit.defaults.title, newTitle)
	unit.defaults.titleShort = Utils.getNoNil(unit.defaults.titleShort, newTitle)

	unit.title = newTitle
	unit.titleShort = newTitleShort
	unit.hasUserTitle = newTitle ~= l10n_title
	unit.hasUserTitleShort = newTitleShort ~= l10n_titleShort
end

function UnitExtension:getUnitType(unitType)
-- Returns a unit type table
	-- [unitType]: Target unit type (name or index)

	unitType = HMUtils.upperCase(unitType)
	return Utils.getNoNil(self.unitTypesByName[unitType], self.unitTypesByIndex[unitType])
end

function UnitExtension:getUnitSet(unitSet)
-- Returns a unit set table (i.e. VEHICLE1, VEHICLE2, ANIMAL, PRICE)
	-- [unitSet]: Target unit set (name or index)

	unitSet = HMUtils.upperCase(unitSet)
	return Utils.getNoNil(self.unitSetsByName[unitSet], self.unitSetsByIndex[unitSet])
end

function UnitExtension:getUnit(unit)
-- Returns a unit table
	-- [unit]: 	Target unit (name or index)

	unit = HMUtils.upperCase(unit)
	return Utils.getNoNil(self.unitsByName[unit], self.unitsByIndex[unit])
end

function UnitExtension:isFillTypeUnit(unit)
-- Returns true if unit is a fillType unit (volume, weight, or piece)
	-- [unit]: Target unit (name or index)

	unit = self:getUnit(unit)

	if unit	and (unit.type == self.unitType.VOLUME or unit.type == self.unitType.WEIGHT	or unit.type == self.unitType.PIECE) then
		return true
	end

	return false
end

-- [FillType Group Functions] **********************************************************************************************************************

function UnitExtension:registerFillTypeGroup(name, title, isLocked, ...)
-- Registers and returns a new fillType group
	-- [name]:		FillType group name
	-- [title]:		FillType group title
	-- [isLocked]:	Prevent fillType group units from being changed
	-- [...]		Units passed to setFillTypeGroupUnits function (any) > Optional

	isLocked = Utils.getNoNil(isLocked, false)

	if HMUtils.argIsInvalid(HMUtils.isString(name), "name")
	or HMUtils.argIsInvalid(HMUtils.isBoolean(isLocked), "isLocked")
	then
		return
	end

	local groupInfo = self:getFillTypeGroup(name)
	if HMUtils.errorOnTrue(groupInfo, self:getMsg("FILLTYPE_GROUP_EXISTS"), name) then
		return groupInfo
	end

	groupInfo = {
		name	  = HMUtils.upperCase(name),
		locked	  = isLocked,
		isUser	  = false, -- this is set in the load xml on registration
		defaults  = {},
		units	  = {}
	}
	table.insert(self.fillTypeGroupsByIndex, groupInfo)
	groupInfo.index = #self.fillTypeGroupsByIndex
	self.fillTypeGroupsByName[groupInfo.name] = groupInfo
	self.fillTypeGroup[groupInfo.name] = groupInfo.index

	self:setFillTypeGroupTitle(groupInfo.name, title)
	self:setFillTypeGroupUnits(groupInfo.name, nil, ...)

	HMUtils.displayMsg("FillType group %q registered successfully", groupInfo.name)
	return groupInfo
end

function UnitExtension:setFillTypeGroupTitle(group, title)
-- Sets specified fillType group title
	-- [group]: Target fillType group (name or index)
	-- [title]: FillType group title (string)

	local fillTypeGroup = self:getFillTypeGroup(group)
	if HMUtils.argIsInvalid(fillTypeGroup, "group") then
		return
	end

	local l10n_title = UnitExtension.L10N_PREFIX.."fillTypeGroup_"..HMUtils.lowerCase(fillTypeGroup.name)
	l10n_title = HMUtils.getText(l10n_title)

	local oldTitle = Utils.getNoNil(fillTypeGroup.title, l10n_title)
	local newTitle = Utils.getNoNil(title, oldTitle)

	if HMUtils.argIsInvalid(HMUtils.isString(newTitle), "title") then
		newTitle = oldTitle
	end

	fillTypeGroup.defaults.title = Utils.getNoNil(fillTypeGroup.defaults.title, newTitle)
	fillTypeGroup.title = newTitle
	fillTypeGroup.hasUserTitle = newTitle ~= l10n_title
end

function UnitExtension:setFillTypeGroupUnits(group, set, ...)
-- Sets specified fillType group unit values
	-- [group]: Target fillType group (name or index)
	-- [set]:	Target unit set (name or index) > Optional
	-- [...]: 	Unit(s) to assign (name or index or table) > Default = Set all values to "LITER"

	local fillTypeGroup = self:getFillTypeGroup(group)
	local unitSet = self:getUnitSet(set)
	local unitList = HMUtils.pack(...)

	if HMUtils.argIsInvalid(fillTypeGroup, "group") then
		return
	end

	fillTypeGroup.defaults.units = Utils.getNoNil(fillTypeGroup.defaults.units, {})
	local defaultUnits = fillTypeGroup.defaults.units

	if unitSet then
		HMUtils.errorOnTrue(unitList.n > 1, self:getMsg("INVALID_FUNCTION_CALL"))
		local oldUnit = Utils.getNoNil(fillTypeGroup.units[unitSet.index], self.unit.LITER)
		local newUnit = Utils.getNoNil(unitList[1], oldUnit)
		if HMUtils.argIsInvalid(self:isFillTypeUnit(newUnit), unitSet.name) then
			newUnit = oldUnit
		end
		newUnit = self:getUnit(newUnit)
		defaultUnits[unitSet.index] = Utils.getNoNil(defaultUnits[unitSet.index], newUnit.index)
		fillTypeGroup.units[unitSet.index] = newUnit.index
	else
		if HMUtils.argIsInvalid(set == nil, "set") then
			return
		end

		if HMUtils.isTable(unitList[1]) then
			HMUtils.errorOnTrue(unitList.n > 1, self:getMsg("INVALID_FUNCTION_CALL"))
			unitList = unitList[1]
		end

		local baseUnit = Utils.getNoNil(unitList[1], self.unit.LITER)
		for i, unitSet in ipairs(self.unitSetsByIndex) do
			local oldUnit = Utils.getNoNil(fillTypeGroup.units[i], baseUnit)
			local newUnit = Utils.getNoNil(unitList[i], oldUnit)
			if HMUtils.argIsInvalid(self:isFillTypeUnit(newUnit), unitSet.name) then
				newUnit = oldUnit
			end
			newUnit = self:getUnit(newUnit)
			defaultUnits[i] = Utils.getNoNil(defaultUnits[i], newUnit.index)
			fillTypeGroup.units[i] = newUnit.index
		end
	end
end

function UnitExtension:getFillTypeGroup(group)
-- Returns a fillType group table
	-- [group]: FillType group to retrieve (name or index)

	group = HMUtils.upperCase(group)
	return Utils.getNoNil(self.fillTypeGroupsByName[group], self.fillTypeGroupsByIndex[group])
end

function UnitExtension:getFillTypeGroupUnit(group, set)
-- Returns a fillType group unit table
	-- [group]: Target fillType group (name or index)
	-- [set]: 	Target unit set (name or index)

	local fillTypeGroup = self:getFillTypeGroup(group)

	if HMUtils.isString(set) then
		set = HMUtils.upperCase(set)
		local setIdx = self.currentUnitSet[set]
		if setIdx then
			set = set..tostring(setIdx)
		end
	end

	local unitSet = self:getUnitSet(set)

	if not (fillTypeGroup and unitSet) then
		return
	end

	return self:getUnit(fillTypeGroup.units[unitSet.index])
end

-- [FillType Functions] ****************************************************************************************************************************

function UnitExtension:registerFillType(fillType, group, unScaleMass)
-- Registers and returns an extended fillType table
	-- [fillType]: 		FillType to register (name or index)
	-- [group]:	   		FillType group to set (name or index) > See UnitExtension:setFillTypeGroup()
	-- [unScaleMass]:	Unscale filltype mass using Giants FillType.MASS_SCALE value (0.5 as of v1.4.0.0) > Default = false

	fillType = HMUtils.getFillType(fillType)
	unScaleMass = Utils.getNoNil(unScaleMass, false)

	if HMUtils.argIsInvalid(fillType, "fillType")
	or HMUtils.argIsInvalid(HMUtils.isBoolean(unScaleMass), "unScaleMass")
	then
		return
	end

	local fillTypeInfo = self:getFillType(fillType.name)
	if HMUtils.errorOnTrue(fillTypeInfo, self:getMsg("FILLTYPE_EXISTS"), fillType.name) then
		return fillTypeInfo
	end

	fillTypeInfo = {
		parent			 = fillType, -- used for comparisons
		name 			 = fillType.name,
		index 			 = fillType.index,
		title			 = fillType.title,
		massPerLiter	 = fillType.massPerLiter,
		defaults		 = {}
	}

	if unScaleMass == true then
		fillTypeInfo.massPerLiter = fillTypeInfo.massPerLiter / FillTypeManager.MASS_SCALE
	end

	fillTypeInfo.defaults.oldMassPerLiter = fillType.massPerLiter  -- to determine if Giants fillType mass has changed
	fillTypeInfo.defaults.massPerLiter = fillTypeInfo.massPerLiter -- to determine if UnitExtension fillType mass has changed

	self.fillTypesByIndex[fillTypeInfo.index] = fillTypeInfo
	self.fillTypesByName[fillTypeInfo.name] = fillTypeInfo
	self.fillType[fillTypeInfo.name] = fillTypeInfo.index

	self:setFillTypeGroup(fillType.name, group)

	HMUtils.displayMsg("FillType %q registered successfully", fillTypeInfo.name)
	return fillTypeInfo
end

function UnitExtension:setFillTypeGroup(fillType, group)
-- Assigns a fillType to a fillType group
	-- [fillType]:  Target fillType (name or index)
	-- [group]:		Target fillType group (name or index)

	fillType = self:getFillType(fillType)

	if HMUtils.argIsInvalid(fillType, "fillType") then
		return
	end

	local oldGroup = Utils.getNoNil(fillType.group, self.fillTypeGroup.MISC)
	local newGroup = Utils.getNoNil(group, oldGroup)

	newGroup = self:getFillTypeGroup(newGroup)

	if HMUtils.argIsInvalid(HMUtils.isTable(newGroup), "group") then
		newGroup = self:getFillTypeGroup(oldGroup)
	end

	fillType.defaults.group = Utils.getNoNil(fillType.defaults.group, newGroup.index)
	fillType.group = newGroup.index
	fillType.hidden = fillType.group == self.fillTypeGroup.MISC
end

function UnitExtension:setFillTypeMass(fillType, mass)
-- Sets a fillType display mass
	-- [fillType]: 	Target fillType (name or index)
	-- [mass]:		Display mass to set > in mt/l

	fillType = self:getFillType(fillType)

	if HMUtils.argIsInvalid(fillType, "fillType")
	or HMUtils.argIsInvalid(mass == nil or HMUtils.isNumber(mass), "mass")
	then
		return
	end

	fillType.massPerLiter = Utils.getNoNil(mass, fillType.massPerLiter)
end

function UnitExtension:setFillTypeShowOnPriceTable(fillType, showOnPriceTable)
-- Sets a fillType showOnPriceValue setting
	-- [fillType]: 			Target fillType (name or index)
	-- [showOnPriceTable]: 	Show fillType in prices menu

	fillType = self:getFillType(fillType)

	if HMUtils.argIsInvalid(fillType, "fillType")
	or HMUtils.argIsInvalid(showOnPriceTable == nil or HMUtils.isBoolean(showOnPriceTable), "showOnPriceTable")
	then
		return
	end

	local hasHudOverlay = false
	local hudOverlayFilename = fillType.parent.hudOverlayFilename
	local hudOverlayFilenameSmall = fillType.parent.hudOverlayFilenameSmall
	if HMUtils.isString(hudOverlayFilename) and HMUtils.isString(hudOverlayFilenameSmall) then
		hasHudOverlay = true
	end

	if fillType.parent.showOnPriceTable ~= nil and hasHudOverlay then
		fillType.defaults.showOnPriceTable = fillType.parent.showOnPriceTable
		fillType.parent.showOnPriceTable = Utils.getNoNil(showOnPriceTable, fillType.parent.showOnPriceTable)
	else
		HMUtils.debugMsg(self, HMUtils.DEBUG.TESTING, "FillType %q cannot be shown on the prices menu", fillType.name)
	end
end

function UnitExtension:getFillType(fillType)
-- Returns both an internal fillType and default fillType table
	-- [fillType]: Target fillType (name or index)

	fillType = HMUtils.upperCase(fillType)
	return Utils.getNoNil(self.fillTypesByName[fillType], self.fillTypesByIndex[fillType])
end

function UnitExtension:getFillTypeUnit(fillType, unitSet)
-- Returns a fillType unit table and conversion factor
	-- [fillType]:  Target fillType (name or index)
	-- [unitSet]: Target unit group (name or index)

	fillType = self:getFillType(fillType)

	if HMUtils.argIsInvalid(fillType, "fillType") then
		return
	end

	local unit = self:getFillTypeGroupUnit(fillType.group, unitSet) -- errors handed by getFillTypeGroupUnit()
	if HMUtils.errorOnFalse(unit, self:getMsg("INVALID_UNIT")) then
		return
	end

	local factor = unit.factor

	if unit.type == self.unitType.WEIGHT then
		factor = factor * math.max(0.000001, fillType.massPerLiter) -- minimum 1 kg/m3
	end

	return unit, factor
end

-- [Pseudo HarvestMaster] #############################################################################################################################################################################

-- [Mission00 Overrides] ****************************************************************************************************************************

function UnitExtension.Mission00.new(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Mission00.new")

	--g_ucLite:onPreInitialize(...)

	local ret = HMUtils.pack(savedFunc(...))

	g_ucLite.isServer = g_server ~= nil
	g_ucLite.isClient = g_client ~= nil

	--g_ucLite:onInitMap(...)
	--g_ucLite:onPostInitMap(...)

	return HMUtils.unpack(ret)
end
HMUtils.setFunction(g_ucLite, "Mission00", "new")

function UnitExtension.Mission00:load(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Mission00.load")

	--g_ucLite:onPreLoadMap(...)

	local ret = HMUtils.pack(savedFunc(self, ...))

	--g_ucLite:onLoadMap(...)
	--g_ucLite:onPostLoadMap(...)

	return HMUtils.unpack(ret)
end
HMUtils.setFunction(g_ucLite, "Mission00", "load")

function UnitExtension.Mission00:loadMission00Finished(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Mission00.loadMission00Finished")

	--g_ucLite:onPreLoadMapFinished(...)

	local ret = HMUtils.pack(savedFunc(self, ...))

	g_ucLite:onLoadMapFinished(...)
	--g_ucLite:onPostLoadMapFinished(...)

	return HMUtils.unpack(ret)
end
HMUtils.setFunction(g_ucLite, "Mission00", "loadMission00Finished")

-- Application Start
g_ucLite:onInitialize()