--[[

Game:	Farming Simulator 19
Title:	UnitConvert Lite (Vehicles)
Author:	ThundR

--]]

-- [Declarations] ***********************************************************************************************************************************

UnitExtension.Vehicle = {}

UnitExtension.Combine = {}
UnitExtension.Combine.YIELD_SEND_BITS = 16

UnitExtension.Cutter = {}
UnitExtension.Drivable = {}
UnitExtension.WorkArea = {}
UnitExtension.FillUnit = {}

-- [Initialization] *********************************************************************************************************************************

function UnitExtension:initializeVehicles()
-- Initializes Vehicles

	-- Vehicle overrides
	HMUtils.setFunction(self, "Vehicle", "loadFinished")

	-- Drivable overrides
	HMUtils.setFunction(self, "Drivable", "onLoad")
	HMUtils.setFunction(self, "Drivable", "onRegisterActionEvents")
	HMUtils.setFunction(self, "Drivable", "registerFunctions")

	-- WorkArea additional events
	HMUtils.setFunction(self, "WorkArea", "onRegisterActionEvents", false)
	HMUtils.setFunction(self, "WorkArea", "onReadUpdateStream", false)
	HMUtils.setFunction(self, "WorkArea", "onWriteUpdateStream", false)
	HMUtils.setFunction(self, "WorkArea", "saveToXMLFile", false)

	-- WorkArea overrides
	HMUtils.setFunction(self, "WorkArea", "onLoad")
	HMUtils.setFunction(self, "WorkArea", "onUpdateTick")
	HMUtils.setFunction(self, "WorkArea", "registerFunctions")
	HMUtils.setFunction(self, "WorkArea", "registerEventListeners")

	-- Cutter overrides
	HMUtils.setFunction(self, "Cutter", "onEndWorkAreaProcessing")

	-- Combine overrides
	HMUtils.setFunction(self, "Combine", "onUpdateTick")
	HMUtils.setFunction(self, "Combine", "onReadUpdateStream")
	HMUtils.setFunction(self, "Combine", "onWriteUpdateStream")
	HMUtils.setFunction(self, "Combine", "registerFunctions")
	HMUtils.setFunction(self, "Combine", "addCutterArea")

	-- FillUnit overrides
	HMUtils.setFunction(self, "FillUnit", "getFillLevelInformation")
end

-- [VEHICLE] ##########################################################################################################################################################################################

-- [Vehicle Overrides] ******************************************************************************************************************************

function UnitExtension.Vehicle:loadFinished(...)
-- Do certain things here because some features may apply to multiple specializations

	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Vehicle.loadFinished")

	self.unitExtension = {}

	local ret = HMUtils.pack(savedFunc(self, ...))

	-- Worked Area Display
	if self.spec_workArea then
		self:initializeWorkAreaParameters()
	end

	-- Yield Display
	if self.spec_combine then
		self:clearYield()
	end

	return HMUtils.unpack(ret)
end

-- [DRIVABLE] #########################################################################################################################################################################################

-- [Drivable Additional Functions] ******************************************************************************************************************

function UnitExtension.Drivable.toggleUnitSet(self)
-- Toggles the current vehicle unit set

	local unitSetIdx = self.unitExtension.currentUnitSet
	unitSetIdx = unitSetIdx + 1

	if unitSetIdx > g_ucLite.numUnitSets.VEHICLE then
		unitSetIdx = 1
	end

	self.unitExtension.currentUnitSet = unitSetIdx
end

function UnitExtension.Drivable:getUnitSet()
-- Returns current unit set name (i.e. VEHICLE1, VEHICLE2)

	return "VEHICLE"..tostring(self.unitExtension.currentUnitSet)
end

-- [Drivable Overrides] *****************************************************************************************************************************

function UnitExtension.Drivable.registerFunctions(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Drivable.registerFunctions")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(...))

	local vehicleType = HMUtils.getRealArg(args, 1)

	SpecializationUtil.registerFunction(vehicleType, "getUnitSet", UnitExtension.Drivable.getUnitSet)

	return HMUtils.unpack(ret)
end

function UnitExtension.Drivable:onLoad(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Drivable.onLoad")
	local ret = HMUtils.pack(savedFunc(self, ...))

	self.unitExtension.currentUnitSet = 1

	return HMUtils.unpack(ret)
end

function UnitExtension.Drivable:onRegisterActionEvents(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Drivable.onRegisterActionEvents")
	local ret = HMUtils.pack(savedFunc(self, ...))

	if self.isClient then
		local spec = self.spec_drivable

		local entered = true
		if self.getIsEntered ~= nil then
			entered = self:getIsEntered()
		end

		if self:getIsActiveForInput(true, true) and entered then
			local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.UE_ToggleUnitSet, self, UnitExtension.Drivable.toggleUnitSet, false, true, false, true)
			g_inputBinding:setActionEventText(actionEventId, g_ucLite:getLocalText("action_toggleUnitSet"))
			g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_LOW)
			self.unitExtension.toggleDisplayUnitEventId = actionEventId
		end
	end

	return HMUtils.unpack(ret)
end

-- [WORKAREA] #########################################################################################################################################################################################

-- [ResetWorkArea Event] ****************************************************************************************************************************

ResetWorkedAreaEvent = {}
ResetWorkedAreaEvent_mt = Class(ResetWorkedAreaEvent, Event)

InitEventClass(ResetWorkedAreaEvent, "ResetWorkedAreaEvent")

function ResetWorkedAreaEvent:emptyNew()
	local self = Event:new(ResetWorkedAreaEvent_mt)
	return self
end

function ResetWorkedAreaEvent:new(vehicle)
	local self = ResetWorkedAreaEvent:emptyNew()
	self.vehicle = vehicle
	return self
end

function ResetWorkedAreaEvent:readStream(streamId, connection)
    self.vehicle = NetworkUtil.readNodeObject(streamId)
    self:run(connection)
end

function ResetWorkedAreaEvent:writeStream(streamId, connection)
    NetworkUtil.writeNodeObject(streamId, self.vehicle)
end

function ResetWorkedAreaEvent:run(connection)
    if self.vehicle.unitExtension.totalWorkedArea then
		self.vehicle.unitExtension.totalWorkedArea = 0
	end
end

-- [WorkArea Additional Functions] ******************************************************************************************************************

function UnitExtension.WorkArea.toggleWorkedAreaDisplay(self)
-- Toggles showing the worked area display on the vehicle hud

	self.unitExtension.showWorkedArea = not self.unitExtension.showWorkedArea
end

function UnitExtension.WorkArea.resetTotalWorkedArea(self)
-- Resets worked area counter to zero

	if not self.isServer then
		g_client:getServerConnection():sendEvent(ResetWorkedAreaEvent:new(self))
	else
		self.unitExtension.totalWorkedArea = 0
	end
end

function UnitExtension.WorkArea:initializeWorkAreaParameters()
-- Initializes vehicle work area parameters

	if self.isServer then
		self.unitExtension.workAreaParameters = {}

		for key, value in pairs(self) do
			key = tostring(key)
			if string.find(key, "^spec_") then
				local workAreaParameters = value.workAreaParameters
				if HMUtils.isTable(workAreaParameters) then
					table.insert(self.unitExtension.workAreaParameters, workAreaParameters)
				end
			end
		end
	end
end

function UnitExtension.WorkArea:updateTotalWorkedArea()
-- Updates and returns total worked area of vehicle

	if self.isServer then
		local workAreaParameters = self.unitExtension.workAreaParameters
		local totalWorkedArea = Utils.getNoNil(self.unitExtension.totalWorkedArea, 0)

		for idx, parameter in ipairs(workAreaParameters) do
			local lastStatsArea = parameter.lastStatsArea
			if HMUtils.isNumber(lastStatsArea) then
				totalWorkedArea = totalWorkedArea + lastStatsArea
			else
				table.remove(workAreaParameters, idx)
			end
		end

		if #workAreaParameters > 0 then
			self.unitExtension.totalWorkedArea = totalWorkedArea
		else
			self.unitExtension.totalWorkedArea = nil
		end
	end

	if self.unitExtension.totalWorkedArea == nil then
		self.unitExtension.showWorkedArea = false
	end
end

-- [WorkArea Additional Events] *********************************************************************************************************************

function UnitExtension.WorkArea:onRegisterActionEvents(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "WorkArea.onRegisterActionEvents")
	local ret = HMUtils.pack(savedFunc(self, ...))

	if self.isClient then
		local spec = self.spec_workArea
		self:clearActionEventsTable(spec.actionEvents)

		if self:getIsActiveForInput(false, true) then
			local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.UE_VehicleFunction1, self, UnitExtension.WorkArea.toggleWorkedAreaDisplay, false, true, false, true)
			g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_LOW)
			g_inputBinding:setActionEventTextVisibility(actionEventId, false)
			self.unitExtension.toggleWorkedAreaEventId = actionEventId

			local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.UE_VehicleFunction2, self, UnitExtension.WorkArea.resetTotalWorkedArea, false, true, false, true)
			g_inputBinding:setActionEventText(actionEventId, g_ucLite:getLocalText("action_resetWorkedArea"))
			g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_LOW)
			g_inputBinding:setActionEventTextVisibility(actionEventId, false)
			self.unitExtension.resetWorkedAreaEventId = actionEventId
		end
	end

	return HMUtils.unpack(ret)
end

function UnitExtension.WorkArea:onReadUpdateStream(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "WorkArea.onReadUpdateStream")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(self, ...))

	local streamId = HMUtils.getRealArg(args, 1)
	local connection = HMUtils.getRealArg(args, 3)

	if connection:getIsServer() then
		local totalWorkedArea = streamReadBool(streamId)
		if totalWorkedArea then
			self.unitExtension.totalWorkedArea = streamReadFloat32(streamId)
		end
	end

	return HMUtils.unpack(ret)
end

function UnitExtension.WorkArea:onWriteUpdateStream(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "WorkArea.onWriteUpdateStream")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(self, ...))

	local streamId = HMUtils.getRealArg(args, 1)
	local connection = HMUtils.getRealArg(args, 2)

	if not connection:getIsServer() then
		streamWriteBool(streamId, self.unitExtension.totalWorkedArea ~= nil)
		if self.unitExtension.totalWorkedArea then
			streamWriteFloat32(streamId, self.unitExtension.totalWorkedArea)
		end
	end

	return HMUtils.unpack(ret)
end

function UnitExtension.WorkArea:saveToXMLFile(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "WorkArea.saveToXMLFile")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(self, ...))

	local spec = self.spec_workArea

	local xmlFile = HMUtils.getRealArg(args, 1)
	local key = HMUtils.getRealArg(args, 2)

	if self.unitExtension.totalWorkedArea ~= nil then
		setXMLFloat(xmlFile, key.."#totalWorkedArea", self.unitExtension.totalWorkedArea)
		setXMLBool(xmlFile, key.."#showCounter", self.unitExtension.showWorkedArea)
	end

	return HMUtils.unpack(ret)
end

-- [WorkArea Overrides] *****************************************************************************************************************************

function UnitExtension.WorkArea.registerFunctions(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "WorkArea.registerFunctions")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(...))

	local vehicleType = HMUtils.getRealArg(args, 1)

	SpecializationUtil.registerFunction(vehicleType, "initializeWorkAreaParameters", UnitExtension.WorkArea.initializeWorkAreaParameters)
	SpecializationUtil.registerFunction(vehicleType, "updateTotalWorkedArea", UnitExtension.WorkArea.updateTotalWorkedArea)

	return HMUtils.unpack(ret)
end

function UnitExtension.WorkArea.registerEventListeners(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "WorkArea.registerEventListeners")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(...))

	local vehicleType = HMUtils.getRealArg(args, 1)

	SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", WorkArea)
	SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", WorkArea)
	SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", WorkArea)

	return HMUtils.unpack(ret)
end

function UnitExtension.WorkArea:onLoad(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "WorkArea.onLoad")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(self, ...))

	self.unitExtension.showWorkedArea = false

	local savegame = HMUtils.getRealArg(args, 1)
	if savegame then
		local xmlFile = savegame.xmlFile
		local key = savegame.key..".workArea"

		if hasXMLProperty(xmlFile, key) then
			if self.isServer then
				self.unitExtension.totalWorkedArea = getXMLFloat(xmlFile, key.."#totalWorkedArea")
			end
			self.unitExtension.showWorkedArea = Utils.getNoNil(getXMLBool(xmlFile, key.."#showCounter"), self.unitExtension.showWorkedArea)
		end
	end

	return HMUtils.unpack(ret)
end

function UnitExtension.WorkArea:onUpdateTick(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "WorkArea.onUpdateTick")
	local ret = HMUtils.pack(savedFunc(self, ...))

	self:updateTotalWorkedArea()

	local allowToggle = false
	local allowReset = false

	if self:getIsActiveForInput(false, true) then
		if self.unitExtension.totalWorkedArea then
			if self.operatingTime ~= nil then
				if self.unitExtension.showWorkedArea == true then
					g_inputBinding:setActionEventText(self.unitExtension.toggleWorkedAreaEventId, g_ucLite:getLocalText("action_showOperatingTime"))
				else
					g_inputBinding:setActionEventText(self.unitExtension.toggleWorkedAreaEventId, g_ucLite:getLocalText("action_showWorkedArea"))
				end
				allowToggle = true
			else
				self.unitExtension.showWorkedArea = true
			end

			if self.unitExtension.showWorkedArea == true and self.unitExtension.totalWorkedArea > 0 then
				allowReset = true
			end
		end
	end

	g_inputBinding:setActionEventTextVisibility(self.unitExtension.toggleWorkedAreaEventId, allowToggle)
	g_inputBinding:setActionEventTextVisibility(self.unitExtension.resetWorkedAreaEventId, allowReset)
	g_inputBinding:setActionEventActive(self.unitExtension.toggleWorkedAreaEventId, allowToggle)
	g_inputBinding:setActionEventActive(self.unitExtension.resetWorkedAreaEventId, allowReset)

	return HMUtils.unpack(ret)
end

-- [CUTTER] ###########################################################################################################################################################################################

-- [Cutter Overrides] *******************************************************************************************************************************

function UnitExtension.Cutter:onEndWorkAreaProcessing(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Cutter.onEndWorkAreaProcessing")

	local spec = self.spec_cutter
	local combineVehicle = spec.workAreaParameters.combineVehicle

	if combineVehicle then
		local combineSpec = combineVehicle.spec_combine
		if combineSpec then
			combineSpec.lastValidCutter = self
		end
	end

	local ret = HMUtils.pack(savedFunc(self, ...))
	return HMUtils.unpack(ret)
end

-- [COMBINE] ##########################################################################################################################################################################################

-- [Combine Additional Functions] *******************************************************************************************************************

function UnitExtension.Combine:clearYield()
-- Clears the vehicle yield tables

	self.unitExtension.yieldFillTypes = {}
	self.unitExtension.yieldFillTypeMapping = {}
end

function UnitExtension.Combine:setYield(fillType, yield)
-- Sets the yield value(s) of current vehicle
	-- [fillType]: Yield fill type (index or name)
	-- [yield]:	   Yield value

	if HMUtils.argIsInvalid(HMUtils.isNumber(yield) and yield >= 0, "yield") then
		return HMUtils.unpack(ret)
	end

	fillType = g_ucLite:getFillType(fillType)
	if HMUtils.argIsInvalid(fillType, "fillType") then
		return HMUtils.unpack(ret)
	end

	local maxSentYield = (2 ^ UnitExtension.Combine.YIELD_SEND_BITS) - 1
	yield = MathUtil.clamp(MathUtil.round(yield * 100), 0, maxSentYield) / 100

	local yieldFillTypes = self.unitExtension.yieldFillTypes
	if yieldFillTypes[fillType.index] then
		yieldFillTypes[fillType.index] = (yieldFillTypes[fillType.index] + yield) / 2 -- average
	else
		table.insert(self.unitExtension.yieldFillTypeMapping, fillType.index)
		yieldFillTypes[fillType.index] = yield
	end
end

-- [Combine Overrides] ******************************************************************************************************************************

function UnitExtension.Combine.registerFunctions(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Combine.registerFunctions")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(...))

	local vehicleType = HMUtils.getRealArg(args, 1)

	SpecializationUtil.registerFunction(vehicleType, "setYield", UnitExtension.Combine.setYield)
	SpecializationUtil.registerFunction(vehicleType, "clearYield", UnitExtension.Combine.clearYield)

	return HMUtils.unpack(ret)
end

function UnitExtension.Combine:addCutterArea(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Combine.addCutterArea")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(self, ...))

	local spec = self.spec_combine

	if args.n > 0 and ret.n > 0 then
		if self.isServer then
			local fruitType = HMUtils.getFruitType(HMUtils.getRealArg(args, 3))
			local fillType = HMUtils.getFillType(HMUtils.getRealArg(args, 4))
			local appliedDelta = ret[1]
			local threshedArea = 0

			local lastValidCutter = spec.lastValidCutter

			if fruitType and spec.lastCuttersInputFruitType == fruitType.index
			and fillType and spec.lastCuttersOutputFillType == fillType.index
			and HMUtils.isNumber(appliedDelta) and appliedDelta > 0
			and lastValidCutter
			then
				local cutterSpec = lastValidCutter.spec_cutter
				threshedArea = cutterSpec.workAreaParameters.lastThreshedArea
			end

			if threshedArea > 0 then
				local pixelsPerSqm = g_currentMission:getFruitPixelsToSqm()
				local threshedSqm = threshedArea * pixelsPerSqm
				local yield = math.max(0, appliedDelta / threshedSqm)

				self:setYield(fillType.index, yield)
			end
		end
	end

	return HMUtils.unpack(ret)
end

function UnitExtension.Combine:onUpdateTick(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Combine.onUpdateTick")
	local ret = HMUtils.pack(savedFunc(self, ...))

	local spec = self.spec_combine

	if self.isServer then
		if spec.lastAreaZeroTime > 2000 then -- two seconds
			if #self.unitExtension.yieldFillTypeMapping > 0 then
				self:clearYield()
			end
		end
	end

	return HMUtils.unpack(ret)
end

function UnitExtension.Combine:onReadUpdateStream(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Combine.onReadUpdateStream")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(self, ...))

	local streamId = HMUtils.getRealArg(args, 1)
	local connection = HMUtils.getRealArg(args, 3)

	if connection:getIsServer() then
		self:clearYield()
		local yieldMapping = self.unitExtension.yieldFillTypeMapping
		local yieldFillTypes = self.unitExtension.yieldFillTypes

		local numYieldTypes = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)

		for i = 1, numYieldTypes do
			local fillType = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)
			local yield = streamReadUInt16(streamId) / 100

			table.insert(yieldMapping, fillType)
			yieldFillTypes[fillType] = yield
		end
	end

	return HMUtils.unpack(ret)
end

function UnitExtension.Combine:onWriteUpdateStream(...)
	local savedFunc = HMUtils.getSavedFunction(g_ucLite, "Combine.onWriteUpdateStream")
	local args = HMUtils.pack(...)
	local ret = HMUtils.pack(savedFunc(self, ...))

	local streamId = HMUtils.getRealArg(args, 1)
	local connection = HMUtils.getRealArg(args, 2)

	if not connection:getIsServer() then
		local yieldMapping = self.unitExtension.yieldFillTypeMapping
		local yieldFillTypes = self.unitExtension.yieldFillTypes

		local numYieldTypes = #yieldMapping
		streamWriteUIntN(streamId, numYieldTypes, FillTypeManager.SEND_NUM_BITS)

		for i = 1, numYieldTypes do
			local fillType = yieldMapping[i]
			local yield = yieldFillTypes[fillType] * 100

			streamWriteUIntN(streamId, fillType, FillTypeManager.SEND_NUM_BITS)
			streamWriteUInt16(streamId, yield)
		end
	end

	return HMUtils.unpack(ret)
end

-- [FILLUNIT] #########################################################################################################################################################################################

-- [FillUnit Overrides] *****************************************************************************************************************************

function UnitExtension.FillUnit:getFillLevelInformation(superFunc, ...)
	superFunc(self, ...)

	local args = HMUtils.pack(...)
	local fillLevelInformations = HMUtils.getRealArg(args, 1)
	local spec = self.spec_fillUnit

	for _,fillUnit in pairs(spec.fillUnits) do
        if fillUnit.capacity > 0 and fillUnit.showOnHud then
			local realFillType = fillUnit.fillType
            local fillType = realFillType

            if fillUnit.fillTypeToDisplay ~= FillType.UNKNOWN then
                fillType = fillUnit.fillTypeToDisplay
            end

            local fillLevel = fillUnit.fillLevel
            if fillUnit.fillLevelToDisplay ~= nil then
                fillLevel = fillUnit.fillLevelToDisplay
            end

            local added = false

            for _,fillLevelInformation in pairs(fillLevelInformations) do
                if fillLevelInformation.realFillType == realFillType then
                    fillLevelInformation.fillLevel = fillLevelInformation.fillLevel + fillLevel
                    fillLevelInformation.capacity = fillLevelInformation.capacity + fillUnit.capacity
                    added = true
                    break
                end
            end

            if not added then
                table.insert(fillLevelInformations, {fillType=fillType, realFillType=realFillType, fillLevel=fillLevel, capacity=fillUnit.capacity})
            end
        end
	end
end
